/*
 * c99 complex mathematical functions
 *
 * 2016-08 Kouichi Hirabayashi
 */

#include <stdio.h>
#include <stdlib.h>
#include "Complex.h"
#define PI	3.14159265358979324

COMPLEX Sqrt(x) COMPLEX x;
{
  double c, r, w;
  COMPLEX z;

  c = sqrt(0.5);
  r = cabs(x);
  w = sqrt(r + fabs(creal(x)));
  if (w == 0.0)
	z = (COMPLEX)0.0;
  else if (creal(x) >= 0)
	z = c * w + I * c * cimag(x) / w;
  else
	z = c * fabs(cimag(x)) / w
		+ I * ((cimag(x) >= 0.0) ? c * w : -c * w);
  return z;
}

COMPLEX Exp(x) COMPLEX x;
{
  double a;

  a = exp(creal(x));
  x = a * cos(cimag(x)) + a * sin(cimag(x)) * I;
  return x;
}

COMPLEX Log(x) COMPLEX x;
{
  COMPLEX z;

  z = 0.5 * log(creal(x) * creal(x) + cimag(x) * cimag(x))
	+  atan2(cimag(x), creal(x)) * I;
  return z;
}

COMPLEX Pow(x, y) COMPLEX x, y;
{
  if (x == (COMPLEX)0.0)
	return (COMPLEX)0.0;
  if (y == (COMPLEX)0.0)
	return (COMPLEX)1.0;
  if (y == (COMPLEX)1.0)
	return x;
  return Exp(y* Log(x));
}

// Sin(x) = sin(x.x)*cosh(x.y) + j*cos(x.x)*sinh(x.y)

COMPLEX Sin(x) COMPLEX x;
{
  double e, f;

  e = exp(cimag(x)); f = 1 / e;
  x = 0.5 * cos(creal(x)) * (e - f) + 0.5 * sin(creal(x)) * (e + f) * I;
  return x;
}

// Cos(x) = cos(x.x)*cosh(x.y) + j*sin(x.x)*sinh(x.y)

COMPLEX Cos(x) COMPLEX x;
{
  double e, f;

  e = exp(cimag(x)); f = 1.0 / e;
  x = 0.5 * cos(creal(x)) * (f + e) + 0.5 * sin(creal(x)) * (f - e) * I;
  return x;
}

COMPLEX Tan(x) COMPLEX x;
{
  double e, f, d;

  e = exp(2.0 * cimag(x)); f = 1.0 / e;
  d = cos(2.0 * creal(x)) + 0.5 * (e + f);
  x = sin(2.0 * creal(x)) / d + 0.5 * (e - f) / d * I;
  return x;
}

COMPLEX Sinh(x) COMPLEX x;
{
  double e, f;

  e = exp(creal(x)); f = 1.0 / e;
  x = 0.5 * (e - f) * cos(cimag(x)) + 0.5 * (e + f) * sin(cimag(x)) * I;
  return x;
}

COMPLEX Cosh(x) COMPLEX x;
{
  double e, f;

  e = exp(creal(x)); f = 1.0 / e;
  x = 0.5 * (e + f) * cos(cimag(x)) + 0.5 * (e - f) * sin(cimag(x)) * I;
  return x;
}

COMPLEX Tanh(x) COMPLEX x;
{
  double e, f, d;

//  return Sinh(x) / Cosh(x);
  e = exp(2.0 * creal(x)); f = 1.0 / e;
  d = 0.5 * (e + f) + cos(2.0 * cimag(x));
  x = 0.5 * (e - f) / d + sin(2.0 * cimag(x)) / d;
  return x;

}

COMPLEX Coth(x) COMPLEX x;
{
  return Cosh(x) / Sinh(x);
}

COMPLEX Atanh(x) COMPLEX x;
{
  return Log((1.0 + x) / (1.0 - x)) * 0.5;
}

COMPLEX Acoth(x) COMPLEX x;
{
  return Log((x + 1.0) / (x - 1.0)) * 0.5;
}

COMPLEX Jnd(int n, COMPLEX z)
{
  if (n == 0)
	return -Jn(1, z);
  return (Jn(n - 1, z) - Jn(n + 1, z)) * 0.5;
//  return Jn(n - 1, z) - (COMPLEX)n / z * Jn(n, z);
}

// (d/dx)In(n, z)

COMPLEX Ind(int n, COMPLEX z)
{
  if (n == 0)
	return In(1, z);
  return (In(n - 1, z) + In(n + 1, z)) * 0.5;
//  return In(n - 1, z) - (COMPLEX)n / z * In(n, z);
}

COMPLEX Knd(int n, COMPLEX z)
{
  if(n == 0)
	return -Kn(1, z);
  return (Kn(n - 1, z) + Kn(n + 1, z)) * -0.5;
//  return -Kn(n - 1, z) - (COMPLEX)n / z * Kn(n, z);
}

COMPLEX In(int n, COMPLEX x)
{
  int i;
  COMPLEX z, N, D, S, X, A;

  if (cabs(x) < 60.0) {
	N = D = S = 1.0;
	X = x * x;
	for (i = 1; i <= 100; i++) {
		N *= X;
		D *= 4.0 * (COMPLEX)(i * (n + i));
		A = N / D;
		S += A;
		if (cabs(A) / cabs(S) < 1e-6)
			break;
	}
	if (n == 0)
		return S;
	N = X = 0.5 * x;
	for (i = 2; i <= n; i++)
		N *= X / (COMPLEX)i;
	z = S * N;
  }
  else {
	COMPLEX m, k;

	m = 4.0 * (double)n;
	z = Exp(x) / Sqrt(2.0 * PI * x);
	k = 1.0 - (m - 1.0) / 8.0 / x
		+ (m - 1.0) * (m -9.0) / (128.0 * x * x)
		- (m - 1.0) * (m - 9) * (m - 25.0)
			/ (3072.0 * x * x * x);
	z *= k;
  }
  return z;
}

COMPLEX Jn(int n, COMPLEX x)
{
  int i, k;
  COMPLEX z;
  COMPLEX N, D, S, X, A;

  k = n;
  if (n < 0)
	x = -x;
  if (x == (COMPLEX)0.0)
	return (n == 0) ? (COMPLEX)1.0 : (COMPLEX)0.0;
  if (cabs(x) < 70.0) {
	N = D = S = 1.0;
	X = -x * x;
	for (i = 1; i <= 100; i++) {
		N *= X;
		D *= 4.0 * (double)(i * (n + i));
		A = N / D;
		S += A;
		if (cabs(A) / cabs(S) < 1e-6)
			break;
	}
	if (i >= 100) fprintf(stderr, "! Jn() does not converge (%d).\n", i);
	if (n == 0)
		return S;
	N = X = 0.5 * x;
	for (i = 2; i <= n; i++)
		N *= X / (COMPLEX)i;
	z = S * N;
  }
  else
	z = Sqrt((COMPLEX)2.0 / (PI * x)) * Cos(x - 0.5 * (double)n * PI
		- 0.25 * PI);
  return z;
}

COMPLEX Kn(n, x) int n; COMPLEX x;
{
  int i;
  double g;
  COMPLEX z, z1, z2;

  if (x == (COMPLEX)0.0)
	return INFINITY;
  if (n == 0)
	return K0(x);
  else if (n == 1)
	return K1(x);
  z1 = K0(x);
  z2 = K1(x);
  for (i = 1; i < n; i++) {
	z = z1 - z2 * 2.0 * (double)i / x;
	z1 = z2;
	z2 = z;
  }
  return z;
}

COMPLEX J0(x) COMPLEX x;
{	/* Bessel Function of the First order */
  static double a1[] = { .0002100, -.0039444, .0444479, -.3163866,
	1.2656208, -2.2499997 };
  static double a2[] = { .00014476, -.00072805, .00137237, -.00009512,
	-.00552740, -.00000077 };
  static double a3[] = { .00013558, -.00029333, -.00054125, .00262573,
	-.00003954, -.04166397 };

  int i;
  COMPLEX c, t, y, z;

  c = 3.0;
  if (cabs(x) <= cabs(c)) {
	t = x / c; t = t * t;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a1[i]) * t;
	y += 1.0;
  }
  else {
	t = c / x;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a2[i]) * t;
	y += 0.79788456;
	z = 0.0;
	for (i = 0; i < 6; i++)
		z = (z + a3[i]) * t;
	z -= .78539816;
	z += x;
	y = y * Cos(z) / csqrt(x);
  }
  return y;
}

COMPLEX J1(x) COMPLEX x;
{	/* Bessel Function of the Second order */
  static double a1[] = { .00001109, -.00031761, .00443319, -.03954289,
	.21093573, -.56249985 };
  static double a2[] = { -.00020033, .00113653, -.00249511, .00017105,
	.01659667, .00000156 };
  static double a3[] = { -.00029166, .00079824, .00074348, -.00637879,
	.00005650, .12499612 };
  int i;
  COMPLEX c, t, y, z;

  c = 3.0;
  if (cabs(x) <= cabs(c)) {
	t = x / c; t = t * t;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a1[i]) * t;
	y += 0.5;
	y *= x;
  }
  else {
	t = c / x;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a2[i]) * t;
	y += 0.79788456;
	z = 0.0;
	for (i = 0; i < 6; i++)
		z = (z + a3[i]) * t;
	z -= 2.35619449;
	z += x;
	y *= Cos(z) / csqrt(x);
  }
  return y;
}

COMPLEX I0(x) COMPLEX x;
{	/* Modified Bessel Function of the First order */
  static double a1[] = { 0.0045813, 0.0360768, 0.2659732, 1.2067492,
	3.0899424, 3.5156229 };
  static double a2[] = { 0.00392377, -0.01647633, 0.02635537, -0.02057706,
	0.00916281, -0.00157565, 0.00225319, 0.01328592 };

  int i;
  COMPLEX c, t, y;

  c = 3.75;
  if (cabs(x) <= cabs(c)) {
	t = x / c; t = t * t;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a1[i]) * t;
	y += 1.0;
  }
  else {
	t = c / x;
	y = 0.0;
	for (i = 0; i < 8; i++)
		y = (y + a2[i]) * t;
	y += 0.39894228;
	y *= Exp(x) / csqrt(x);
  }
  return y;
}


COMPLEX I1(x) COMPLEX x;
	{	/* Modified Bessel Function of the First order */
  static double a1[] = { 0.00032411, 0.00301532, 0.02658733, 0.15084934,
	0.51498869, 0.87890594 };
  static double a2[] = { -0.00420059, 0.01787654, -0.02895312, 0.02282967,
	-0.01031555, 0.00163801, -0.00362018, -0.03988024 };
  int i;
  COMPLEX c, t, y;

  c = 3.75;
  if (cabs(x) <= cabs(c)) {
	t = x / c; t = t * t;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a1[i]) * t;
	y += 0.5;
	y *= x;
  }
  else {
	t = c / x;
	y = 0.0;
	for (i = 0; i < 8; i++)
		y = (y + a2[i]) * t;
	y += 0.39894228;
	y *= Exp(x) / csqrt(x);
  }
  return y;
}

COMPLEX K0(x) COMPLEX x;
{	/* Modified Bessel Function of the Second order */
  static double a1[] = { 0.00000740, 0.00010750, 0.00262698, 0.03488590,
	0.23069756, 0.42278420 };
  static double a2[] = { 0.00053208, -0.00251540, 0.00587872, -0.01062446,
	0.02189568, -0.07832358 };
  int i;
  COMPLEX c, t, y;

  c = 2.0;
  if (cabs(x) <= cabs(c)) {
	t = x / c; t = t * t;

	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a1[i]) * t;
	y -= 0.57721566;
	y -=  Log(x * 0.5) * I0(x);
  }
  else {
	t = c / x;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a2[i]) * t;
	y += 1.25331414;
	y *= Exp(-x) / csqrt(x);
  }
  return y;
}

COMPLEX K1(x) COMPLEX x;
{	/* Modified Bessel Function of the Second order */
  int i;
  static double a1[] = { -0.00004686, -0.00110404, -0.01919402, -0.18156897,
	-0.67278579, 0.15443144 };
  static double a2[] = { -0.00068245, 0.00325614, -0.00780353, 0.01504268,
	-0.03655620, 0.23498619 };
  COMPLEX c, t, y;

  c = 2.0;
  if (cabs(x) <= cabs(c)) {
	t = x / c; t = t * t;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a1[i]) * t;
	y += 1.0;
	y = (y + x * Log(x * 0.5) * I1(x)) / x;
  }
  else {
	t = c / x;
	y = 0.0;
	for (i = 0; i < 6; i++)
		y = (y + a2[i]) * t;
	y += 1.25331414;
	y *= Exp(-x) / csqrt(x);
  }
  return y;
}
